@@ -14,6 +14,7 @@
 #define IS_PY3K
 #endif
 
+
 #define MODULE_DOC \
 "Low-level tsinfer interface."
 
@@ -60,6 +61,17 @@ uint64_PyArray_converter(PyObject *in, PyObject **out)
     return NPY_SUCCEED;
 }
 
+static int
+int8_PyArray_converter(PyObject *in, PyObject **out)
+{
+    PyObject *ret = PyArray_FROMANY(in, NPY_INT8, 1, 1, NPY_ARRAY_IN_ARRAY);
+    if (ret == NULL) {
+        return NPY_FAIL;
+    }
+    *out = ret;
+    return NPY_SUCCEED;
+}
+
 /*===================================================================
  * AncestorBuilder
  *===================================================================
@@ -92,13 +104,14 @@ AncestorBuilder_init(AncestorBuilder *self, PyObject *args, PyObject *kwds)
 {
     int ret = -1;
     int err;
-    static char *kwlist[] = {"num_samples", "max_sites", NULL};
-    int num_samples, max_sites;
+    static char *kwlist[] = {"num_samples", "max_sites", "genotype_encoding", "mmap_fd", NULL};
+    int num_samples, max_sites, genotype_encoding;
     int flags = 0;
+    int mmap_fd = -1;
 
     self->builder = NULL;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "ii", kwlist,
-                &num_samples, &max_sites)) {
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "ii|ii", kwlist,
+                &num_samples, &max_sites, &genotype_encoding, &mmap_fd)) {
         goto out;
     }
     self->builder = PyMem_Malloc(sizeof(ancestor_builder_t));
@@ -106,8 +119,9 @@ AncestorBuilder_init(AncestorBuilder *self, PyObject *args, PyObject *kwds)
         PyErr_NoMemory();
         goto out;
     }
+    flags = genotype_encoding;
     Py_BEGIN_ALLOW_THREADS
-    err = ancestor_builder_alloc(self->builder, num_samples, max_sites, flags);
+    err = ancestor_builder_alloc(self->builder, num_samples, max_sites, mmap_fd, flags);
     Py_END_ALLOW_THREADS
     if (err != 0) {
         handle_library_error(err);
@@ -274,34 +288,6 @@ AncestorBuilder_ancestor_descriptors(AncestorBuilder *self)
         }
         PyTuple_SET_ITEM(descriptors, j, py_descriptor);
     }
-    /* j = 0; */
-    /* /1* It's not great that we're breaking encapsulation here and looking */
-    /*  * directly in to the builder's data structures. However, it's quite an */
-    /*  * awkward set of data to communicate, so it seems OK. *1/ */
-    /* for (f = self->builder->num_samples; f > 0; f--) { */
-    /*     for (a = self->builder->frequency_map[f].head; a != NULL; a = a->next) { */
-    /*         map_elem = (pattern_map_t *) a->item; */
-    /*         dims = map_elem->num_sites; */
-    /*         site_array = (PyArrayObject *) PyArray_SimpleNew(1, &dims, NPY_INT32); */
-    /*         if (site_array == NULL) { */
-    /*             goto out; */
-    /*         } */
-    /*         site_array_data = (int32_t *) PyArray_DATA(site_array); */
-    /*         /1* The elements are listed backwards, so reverse them *1/ */
-    /*         k = map_elem->num_sites - 1; */
-    /*         for (s = map_elem->sites; s != NULL; s = s->next) { */
-    /*             site_array_data[k] = (int32_t) s->site; */
-    /*             k--; */
-    /*         } */
-    /*         descriptor = Py_BuildValue("kO", (unsigned long) f, site_array); */
-    /*         if (descriptor == NULL) { */
-    /*             Py_DECREF(site_array); */
-    /*             goto out; */
-    /*         } */
-    /*         PyTuple_SET_ITEM(descriptors, j, descriptor); */
-    /*         j++; */
-    /*     } */
-    /* } */
     ret = descriptors;
     descriptors = NULL;
 out:
@@ -335,13 +321,31 @@ out:
     return ret;
 }
 
+static PyObject *
+AncestorBuilder_get_memsize(AncestorBuilder *self, void *closure)
+{
+    PyObject *ret = NULL;
+
+    if (AncestorBuilder_check_state(self) != 0) {
+        goto out;
+    }
+    ret = Py_BuildValue("k", (unsigned long) ancestor_builder_get_memsize(
+                self->builder));
+out:
+    return ret;
+}
+
+
 static PyMemberDef AncestorBuilder_members[] = {
     {NULL}  /* Sentinel */
 };
 
 static PyGetSetDef AncestorBuilder_getsetters[] = {
     {"num_sites", (getter) AncestorBuilder_get_num_sites, NULL, "The number of sites."},
-    {"num_ancestors", (getter) AncestorBuilder_get_num_ancestors, NULL, "The number of ancestors."},
+    {"num_ancestors", (getter) AncestorBuilder_get_num_ancestors, NULL,
+        "The number of ancestors."},
+    {"mem_size", (getter) AncestorBuilder_get_memsize, NULL,
+        "The number of allocated bytes."},
     {NULL}  /* Sentinel */
 };
 
@@ -429,8 +433,10 @@ TreeSequenceBuilder_init(TreeSequenceBuilder *self, PyObject *args, PyObject *kw
 {
     int ret = -1;
     int err;
-    static char *kwlist[] = {"num_alleles", "max_nodes", "max_edges", NULL};
+    static char *kwlist[] = {"num_alleles", "max_nodes", "max_edges", "ancestral_state", NULL};
     PyArrayObject *num_alleles = NULL;
+    PyArrayObject *ancestral_state = NULL;
+    int8_t *ancestral_state_data = NULL;
     unsigned long max_nodes = 1024;
     unsigned long max_edges = 1024;
     unsigned long num_sites;
@@ -438,13 +444,22 @@ TreeSequenceBuilder_init(TreeSequenceBuilder *self, PyObject *args, PyObject *kw
     int flags = 0;
 
     self->tree_sequence_builder = NULL;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&|kk", kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&|kkO&", kwlist,
             uint64_PyArray_converter, &num_alleles,
-            &max_nodes, &max_edges)) {
+            &max_nodes, &max_edges,
+            int8_PyArray_converter, &ancestral_state)) {
         goto out;
     }
     shape = PyArray_DIMS(num_alleles);
     num_sites = shape[0];
+    if (ancestral_state != NULL) {
+        shape = PyArray_DIMS(ancestral_state);
+        if (shape[0] != (npy_intp) num_sites) {
+            PyErr_SetString(PyExc_ValueError, "ancestral_state array wrong size");
+            goto out;
+        }
+        ancestral_state_data = PyArray_DATA(ancestral_state);
+    }
 
     self->tree_sequence_builder = PyMem_Malloc(sizeof(tree_sequence_builder_t));
     if (self->tree_sequence_builder == NULL) {
@@ -452,7 +467,7 @@ TreeSequenceBuilder_init(TreeSequenceBuilder *self, PyObject *args, PyObject *kw
         goto out;
     }
     err = tree_sequence_builder_alloc(self->tree_sequence_builder,
-            num_sites, PyArray_DATA(num_alleles),
+            num_sites, PyArray_DATA(num_alleles), ancestral_state_data,
             max_nodes, max_edges, flags);
     if (err != 0) {
         handle_library_error(err);
@@ -461,6 +476,7 @@ TreeSequenceBuilder_init(TreeSequenceBuilder *self, PyObject *args, PyObject *kw
     ret = 0;
 out:
     Py_XDECREF(num_alleles);
+    Py_XDECREF(ancestral_state);
     return ret;
 }
 
@@ -1282,23 +1298,24 @@ AncestorMatcher_init(AncestorMatcher *self, PyObject *args, PyObject *kwds)
     int ret = -1;
     int err;
     int extended_checks = 0;
+    int weight_by_n = 1;
     static char *kwlist[] = {"tree_sequence_builder", "recombination",
-        "mismatch", "precision", "extended_checks", NULL};
+        "mismatch", "likelihood_threshold", "extended_checks", "weight_by_n", NULL};
     TreeSequenceBuilder *tree_sequence_builder = NULL;
     PyObject *recombination = NULL;
     PyObject *mismatch = NULL;
     PyArrayObject *recombination_array = NULL;
     PyArrayObject *mismatch_array = NULL;
     npy_intp *shape;
-    unsigned int precision = 22;
+    double likelihood_threshold = DBL_MIN;
     int flags = 0;
 
     self->ancestor_matcher = NULL;
     self->tree_sequence_builder = NULL;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!OO|Ii", kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!OO|dii", kwlist,
                 &TreeSequenceBuilderType, &tree_sequence_builder,
-                &recombination, &mismatch, &precision,
-                &extended_checks)) {
+                &recombination, &mismatch, &likelihood_threshold,
+                &extended_checks, &weight_by_n)) {
         goto out;
     }
     self->tree_sequence_builder = tree_sequence_builder;
@@ -1339,11 +1356,14 @@ AncestorMatcher_init(AncestorMatcher *self, PyObject *args, PyObject *kwds)
     if (extended_checks) {
         flags = TSI_EXTENDED_CHECKS;
     }
+    if (!weight_by_n) {
+        flags |= TSI_DISABLE_WEIGHT_BY_N;
+    }
     err = ancestor_matcher_alloc(self->ancestor_matcher,
             self->tree_sequence_builder->tree_sequence_builder,
             PyArray_DATA(recombination_array),
             PyArray_DATA(mismatch_array),
-            precision, flags);
+            likelihood_threshold, flags);
     if (err != 0) {
         handle_library_error(err);
         goto out;
@@ -1515,12 +1535,20 @@ static PyObject *
 AncestorMatcher_get_total_memory(AncestorMatcher *self, void *closure)
 {
     PyObject *ret = NULL;
+    unsigned long val;
 
     if (AncestorMatcher_check_state(self) != 0) {
         goto out;
     }
-    ret = Py_BuildValue("k", (unsigned long)
-            ancestor_matcher_get_total_memory(self->ancestor_matcher));
+
+#if defined(TSI_NO_ATOMICS)
+    /* Without atomics, return an obviously wrong value */
+     val = (unsigned long) PY_SSIZE_T_MAX;
+#else    
+     val = ancestor_matcher_get_total_memory(self->ancestor_matcher);
+#endif
+    ret = Py_BuildValue("k", val);
+
 out:
     return ret;
 }
