@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2020 University of Oxford
+** Copyright (C) 2020-2023 University of Oxford
 **
 ** This file is part of tsinfer.
 **
@@ -152,8 +152,8 @@ verify_restore_tsb(tree_sequence_builder_t *tsb, tsk_table_collection_t *tables)
         tsb, site, node, derived_state, mut_parent);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
-    ret = tree_sequence_builder_alloc(
-        &other_tsb, tsb->num_sites, tsb->sites.num_alleles, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&other_tsb, tsb->num_sites, tsb->sites.num_alleles,
+        tsb->sites.ancestral_state, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     ret = tree_sequence_builder_restore_nodes(
@@ -331,7 +331,8 @@ initialise_builder(tree_sequence_builder_t *tsb, double oldest_time)
 
 static void
 run_random_data(size_t num_samples, size_t num_sites, int seed,
-    double recombination_rate, double mismatch_rate)
+    double recombination_rate, double mismatch_rate, int ancestor_builder_mmap_fd,
+    int ancestor_builder_options)
 {
     tsk_table_collection_t tables;
     ancestor_builder_t ancestor_builder;
@@ -359,12 +360,13 @@ run_random_data(size_t num_samples, size_t num_sites, int seed,
     }
 
     CU_ASSERT_FATAL(num_samples >= 2);
-    ret = ancestor_builder_alloc(&ancestor_builder, num_samples, num_sites, 0);
+    ret = ancestor_builder_alloc(&ancestor_builder, num_samples, num_sites,
+        ancestor_builder_mmap_fd, ancestor_builder_options);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
-    ret = tree_sequence_builder_alloc(&tsb, num_sites, NULL, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, num_sites, NULL, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = ancestor_matcher_alloc(&ancestor_matcher, &tsb, recombination_rates,
-        mismatch_rates, 6, TSI_EXTENDED_CHECKS);
+        mismatch_rates, 1e-6, TSI_EXTENDED_CHECKS);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     for (j = 0; j < num_sites; j++) {
@@ -376,8 +378,11 @@ run_random_data(size_t num_samples, size_t num_sites, int seed,
         ret = ancestor_builder_add_site(&ancestor_builder, time, genotypes);
         CU_ASSERT_EQUAL_FATAL(ret, 0);
     }
+    /* ancestor_builder_print_state(&ancestor_builder, stdout); */
+    ancestor_builder_print_state(&ancestor_builder, _devnull);
     ret = ancestor_builder_finalise(&ancestor_builder);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
+    ancestor_builder_print_state(&ancestor_builder, _devnull);
 
     initialise_builder(&tsb, ancestor_builder.descriptors[0].time);
 
@@ -447,23 +452,37 @@ test_ancestor_builder_errors(void)
     allele_t genotypes_zeros[4] = { 0, 0, 0, 0 };
     tsk_id_t start, end;
     allele_t haplotype[4];
+    FILE *mmap_file;
 
-    ret = ancestor_builder_alloc(&ancestor_builder, 0, 1, 0);
+    /* Bad file descriptor for mmap FD */
+    ret = ancestor_builder_alloc(&ancestor_builder, 2, 1, -2, 0);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_IO);
+    ancestor_builder_free(&ancestor_builder);
+
+    /* File is opened in the wrong mode */
+    mmap_file = fopen(_tmp_file_name, "w");
+    CU_ASSERT_FATAL(mmap_file != NULL);
+    ret = ancestor_builder_alloc(&ancestor_builder, 2, 1, fileno(mmap_file), 0);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_IO);
+    ancestor_builder_free(&ancestor_builder);
+    fclose(mmap_file);
+
+    ret = ancestor_builder_alloc(&ancestor_builder, 0, 1, -1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_NUM_SAMPLES);
     ancestor_builder_free(&ancestor_builder);
 
-    ret = ancestor_builder_alloc(&ancestor_builder, 1, 1, 0);
+    ret = ancestor_builder_alloc(&ancestor_builder, 1, 1, -1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_NUM_SAMPLES);
     ancestor_builder_free(&ancestor_builder);
 
-    ret = ancestor_builder_alloc(&ancestor_builder, 2, 0, 0);
+    ret = ancestor_builder_alloc(&ancestor_builder, 2, 0, -1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     CU_ASSERT_EQUAL_FATAL(ancestor_builder.num_sites, 0);
     ret = ancestor_builder_add_site(&ancestor_builder, 4, genotypes_ones);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_TOO_MANY_SITES);
     ancestor_builder_free(&ancestor_builder);
 
-    ret = ancestor_builder_alloc(&ancestor_builder, 4, 2, 0);
+    ret = ancestor_builder_alloc(&ancestor_builder, 4, 2, -1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = ancestor_builder_add_site(&ancestor_builder, 4, genotypes_zeros);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -476,7 +495,6 @@ test_ancestor_builder_errors(void)
         ancestor_builder.descriptors[0].num_focal_sites,
         ancestor_builder.descriptors[0].focal_sites, &start, &end, haplotype);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_FOCAL_SITE);
-
     ancestor_builder_free(&ancestor_builder);
 }
 
@@ -489,7 +507,7 @@ test_ancestor_builder_one_site(void)
     allele_t ancestor[1];
     tsk_id_t start, end, focal_sites[1];
 
-    ret = ancestor_builder_alloc(&ancestor_builder, 4, 1, 0);
+    ret = ancestor_builder_alloc(&ancestor_builder, 4, 1, -1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = ancestor_builder_add_site(&ancestor_builder, 4, genotypes);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -528,7 +546,7 @@ test_matching_one_site(void)
     size_t num_edges;
     tsk_id_t *left, *right, *parent;
 
-    ret = tree_sequence_builder_alloc(&tsb, 1, NULL, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, 1, NULL, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = tree_sequence_builder_add_node(&tsb, 2.0, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -536,7 +554,7 @@ test_matching_one_site(void)
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     ret = ancestor_matcher_alloc(
-        &ancestor_matcher, &tsb, &recombination_rate, &mismatch_rate, 12, 0);
+        &ancestor_matcher, &tsb, &recombination_rate, &mismatch_rate, 1e-12, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     ret = ancestor_matcher_find_path(
@@ -555,6 +573,9 @@ test_matching_one_site(void)
     ret = tree_sequence_builder_freeze_indexes(&tsb);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
+    /* Make sure we hit the realloc behaviour on the ancestor matcher */
+    ancestor_matcher.traceback_realloc_size = 10;
+
     ret = ancestor_matcher_find_path(
         &ancestor_matcher, 0, 1, haplotype, match, &num_edges, &left, &right, &parent);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -597,7 +618,7 @@ test_matching_one_site_many_alleles(void)
     allele_t derived_state;
 
     /* Create a linear topology with a mutation over each node */
-    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = tree_sequence_builder_add_node(&tsb, num_nodes, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -623,7 +644,7 @@ test_matching_one_site_many_alleles(void)
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     ret = ancestor_matcher_alloc(
-        &ancestor_matcher, &tsb, &recombination_rate, &mismatch_rate, 12, 0);
+        &ancestor_matcher, &tsb, &recombination_rate, &mismatch_rate, 1e-12, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     haplotype = 0;
@@ -674,7 +695,7 @@ test_matching_errors(void)
     size_t num_edges;
     tsk_id_t *left, *right, *parent;
 
-    ret = tree_sequence_builder_alloc(&tsb, 2, NULL, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, 2, NULL, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = tree_sequence_builder_add_node(&tsb, 2.0, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -682,7 +703,7 @@ test_matching_errors(void)
     CU_ASSERT_EQUAL_FATAL(ret, 0);
 
     ret = ancestor_matcher_alloc(
-        &ancestor_matcher, &tsb, recombination_rate, mismatch_rate, 12, 0);
+        &ancestor_matcher, &tsb, recombination_rate, mismatch_rate, 1e-12, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = ancestor_matcher_find_path(
         &ancestor_matcher, 0, 2, haplotype, match, &num_edges, &left, &right, &parent);
@@ -692,7 +713,7 @@ test_matching_errors(void)
     haplotype[0] = 0;
     mismatch_rate[0] = 1;
     ret = ancestor_matcher_alloc(
-        &ancestor_matcher, &tsb, recombination_rate, mismatch_rate, 12, 0);
+        &ancestor_matcher, &tsb, recombination_rate, mismatch_rate, 1e-12, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
     ret = ancestor_matcher_find_path(
         &ancestor_matcher, 0, 2, haplotype, match, &num_edges, &left, &right, &parent);
@@ -708,26 +729,45 @@ test_tsb_errors(void)
     int ret;
     tree_sequence_builder_t tsb;
     tsk_size_t num_alleles;
+    allele_t ancestral_state;
     tsk_id_t left, right, parent;
     tsk_id_t left_arr[2], right_arr[2], parent_arr[2];
 
     num_alleles = 0;
-    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_NUM_ALLELES);
     tree_sequence_builder_free(&tsb);
 
     num_alleles = 1;
-    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_NUM_ALLELES);
     tree_sequence_builder_free(&tsb);
 
     num_alleles = INT8_MAX + 1;
-    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, 1, 1, 0);
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, NULL, 1, 1, 0);
     CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_NUM_ALLELES);
     tree_sequence_builder_free(&tsb);
 
+    /* Test negative ancestral_state */
     num_alleles = 2;
-    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, 1, 1, 0);
+    ancestral_state = -1;
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, &ancestral_state, 1, 1, 0);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_ANCESTRAL_STATE);
+    tree_sequence_builder_free(&tsb);
+
+    /* Test ancestral_state >= num_alleles */
+    ancestral_state = 2;
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, &ancestral_state, 1, 1, 0);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_ANCESTRAL_STATE);
+    tree_sequence_builder_free(&tsb);
+
+    ancestral_state = 3;
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, &ancestral_state, 1, 1, 0);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_BAD_ANCESTRAL_STATE);
+    tree_sequence_builder_free(&tsb);
+
+    num_alleles = 2;
+    ret = tree_sequence_builder_alloc(&tsb, 1, &num_alleles, NULL, 1, 1, 0);
     /* Add two nodes so we can test adding paths */
     ret = tree_sequence_builder_add_node(&tsb, 2.0, 0);
     CU_ASSERT_EQUAL_FATAL(ret, 0);
@@ -801,8 +841,8 @@ test_random_data_n5_m3(void)
 
     for (seed = 1; seed < 100; seed++) {
         /* printf("seed = %d\n", seed); */
-        run_random_data(5, 3, seed, 1e-3, 1e-20);
-        run_random_data(5, 3, seed, 1e-20, 1e-3);
+        run_random_data(5, 3, seed, 1e-3, 1e-20, -1, 0);
+        run_random_data(5, 3, seed, 1e-20, 1e-3, -1, 0);
     }
 }
 
@@ -810,39 +850,158 @@ static void
 test_random_data_n5_m20(void)
 {
     int seed;
+    size_t j;
+    int options[] = { 0, TSI_GENOTYPE_ENCODING_ONE_BIT };
 
-    for (seed = 1; seed < 10; seed++) {
-        run_random_data(5, 20, seed, 1e-3, 1e-20);
-        run_random_data(5, 20, seed, 1e-20, 1e-3);
+    for (j = 0; j < sizeof(options) / sizeof(*options); j++) {
+        for (seed = 1; seed < 10; seed++) {
+            run_random_data(5, 20, seed, 1e-3, 1e-20, -1, options[j]);
+            run_random_data(5, 20, seed, 1e-20, 1e-3, -1, options[j]);
+        }
     }
 }
 
 static void
 test_random_data_n10_m10(void)
 {
-    run_random_data(10, 10, 43, 1e-3, 1e-20);
-    run_random_data(10, 10, 43, 1e-20, 1e-3);
+    size_t j;
+    int options[] = { 0, TSI_GENOTYPE_ENCODING_ONE_BIT };
+
+    for (j = 0; j < sizeof(options) / sizeof(*options); j++) {
+        run_random_data(10, 10, 43, 1e-3, 1e-20, -1, options[j]);
+        run_random_data(10, 10, 43, 1e-20, 1e-3, -1, options[j]);
+    }
 }
 
 static void
 test_random_data_n10_m100(void)
 {
-    run_random_data(10, 100, 43, 1e-3, 1e-20);
-    run_random_data(10, 100, 43, 1e-20, 1e-3);
+    size_t j;
+    int options[] = { 0, TSI_GENOTYPE_ENCODING_ONE_BIT };
+
+    for (j = 0; j < sizeof(options) / sizeof(*options); j++) {
+        run_random_data(10, 100, 43, 1e-3, 1e-20, -1, options[j]);
+        run_random_data(10, 100, 43, 1e-20, 1e-3, -1, options[j]);
+    }
 }
 
 static void
 test_random_data_n100_m10(void)
 {
-    run_random_data(10, 10, 1243, 1e-3, 1e-20);
-    run_random_data(10, 10, 1243, 1e-20, 1e-3);
+    size_t j;
+    int options[] = { 0, TSI_GENOTYPE_ENCODING_ONE_BIT };
+
+    for (j = 0; j < sizeof(options) / sizeof(*options); j++) {
+        run_random_data(10, 10, 1243, 1e-3, 1e-20, -1, options[j]);
+        run_random_data(10, 10, 1243, 1e-20, 1e-3, -1, options[j]);
+    }
 }
 
 static void
 test_random_data_n100_m100(void)
 {
-    run_random_data(100, 100, 42, 1e-3, 1e-20);
-    run_random_data(100, 100, 42, 1e-20, 1e-3);
+    size_t j;
+    int options[] = { 0, TSI_GENOTYPE_ENCODING_ONE_BIT };
+
+    for (j = 0; j < sizeof(options) / sizeof(*options); j++) {
+        run_random_data(100, 100, 42, 1e-3, 1e-20, -1, options[j]);
+        run_random_data(100, 100, 42, 1e-20, 1e-3, -1, options[j]);
+    }
+}
+
+static void
+test_random_data_ab_mmap(void)
+{
+    size_t j;
+    int options[] = { 0, TSI_GENOTYPE_ENCODING_ONE_BIT };
+    FILE *mmap_file = fopen(_tmp_file_name, "w+");
+
+    CU_ASSERT_FATAL(mmap_file != NULL);
+
+    for (j = 0; j < sizeof(options) / sizeof(*options); j++) {
+        run_random_data(100, 100, 42, 1e-3, 1e-20, fileno(mmap_file), options[j]);
+    }
+    fclose(mmap_file);
+}
+
+static void
+test_packbits_1(void)
+{
+    int ret = 0;
+    allele_t a[] = { 0, 1, 0, 1, 0, 1, 0, 0, 1 };
+    uint8_t b[] = { 42, 1 };
+    uint8_t bitpacked[100];
+    allele_t bitunpacked[100];
+
+    ret = packbits(a, sizeof(a), bitpacked);
+    CU_ASSERT_EQUAL_FATAL(ret, 0);
+    CU_ASSERT_EQUAL(memcmp(b, bitpacked, sizeof(b)), 0);
+    unpackbits(b, sizeof(b), bitunpacked);
+    CU_ASSERT_EQUAL(memcmp(a, bitunpacked, sizeof(a)), 0);
+}
+
+static void
+test_packbits_2(void)
+{
+    int ret = 0;
+    allele_t a[] = { 0, 1, 0, 1, 0, 1, 0, 0 };
+    uint8_t b[] = { 42 };
+    uint8_t bitpacked[100];
+    allele_t bitunpacked[100];
+
+    ret = packbits(a, sizeof(a), bitpacked);
+    CU_ASSERT_EQUAL_FATAL(ret, 0);
+    CU_ASSERT_EQUAL(memcmp(b, bitpacked, sizeof(b)), 0);
+    unpackbits(b, sizeof(b), bitunpacked);
+    CU_ASSERT_EQUAL(memcmp(a, bitunpacked, sizeof(a)), 0);
+}
+
+static void
+test_packbits_3(void)
+{
+    int ret = 0;
+    allele_t a[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };
+    uint8_t b[] = { 255, 127 };
+    uint8_t bitpacked[100];
+    allele_t bitunpacked[100];
+
+    ret = packbits(a, sizeof(a), bitpacked);
+    CU_ASSERT_EQUAL_FATAL(ret, 0);
+    CU_ASSERT_EQUAL(memcmp(b, bitpacked, sizeof(b)), 0);
+    unpackbits(b, sizeof(b), bitunpacked);
+    CU_ASSERT_EQUAL(memcmp(a, bitunpacked, sizeof(a)), 0);
+}
+
+static void
+test_packbits_4(void)
+{
+    int ret = 0;
+    allele_t a[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+    uint8_t b[] = { 0, 0, 0 };
+    uint8_t bitpacked[100];
+    allele_t bitunpacked[100];
+
+    ret = packbits(a, sizeof(a), bitpacked);
+    CU_ASSERT_EQUAL_FATAL(ret, 0);
+    CU_ASSERT_EQUAL(memcmp(b, bitpacked, sizeof(b)), 0);
+    unpackbits(b, sizeof(b), bitunpacked);
+    CU_ASSERT_EQUAL(memcmp(a, bitunpacked, sizeof(a)), 0);
+}
+
+static void
+test_packbits_errors(void)
+{
+    int ret = 0;
+    allele_t a[] = { 0 };
+    uint8_t b[] = { 0 };
+
+    a[0] = -1;
+    ret = packbits(a, sizeof(a), b);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_ONE_BIT_NON_BINARY);
+
+    a[0] = 2;
+    ret = packbits(a, sizeof(a), b);
+    CU_ASSERT_EQUAL_FATAL(ret, TSI_ERR_ONE_BIT_NON_BINARY);
 }
 
 static void
@@ -858,7 +1017,7 @@ test_strerror(void)
         CU_ASSERT(strlen(msg) > 0);
     }
     CU_ASSERT_STRING_EQUAL(
-        tsk_strerror(0), "Normal exit condition. This is not an error!");
+        tsi_strerror(0), "Normal exit condition. This is not an error!");
 }
 
 static int
@@ -929,6 +1088,13 @@ main(int argc, char **argv)
         { "test_random_data_n10_m100", test_random_data_n10_m100 },
         { "test_random_data_n100_m10", test_random_data_n100_m10 },
         { "test_random_data_n100_m100", test_random_data_n100_m100 },
+        { "test_random_data_ab_mmap", test_random_data_ab_mmap },
+
+        { "test_packbits_1", test_packbits_1 },
+        { "test_packbits_2", test_packbits_2 },
+        { "test_packbits_3", test_packbits_3 },
+        { "test_packbits_4", test_packbits_4 },
+        { "test_packbits_errors", test_packbits_errors },
 
         { "test_strerror", test_strerror },
 
